
import math
import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from ggfm.conv.hgt_conv import HGTConv
from gensim.parsing.preprocessing import *


class GPT_GNN(nn.Module):
    
    r"""`"GPT-GNN: Generative Pre-Training of Graph Neural Networks"
    <https://arxiv.org/abs/2006.15437>`_ paper.

    Parameters
    ----------
    gnn: class:`ggfm.models`
        The used GNN model.
    rem_edge_list: dict
        The remaining edge list after sampling.
    attr_decoder: `ggfm.models`
        Attribute decoder.
    neg_samp_num: int
        Maximum number of negative sample for each target node.
        (default: :obj:`1`)
    device: int
        Device
    neg_queue_size: int, optional
        Max size of negetive adaptive embedding queue.
        (default: :obj:`0`)

    """

    def __init__(self, gnn, rem_edge_list, attr_decoder, neg_samp_num, device, neg_queue_size=0):
        super(GPT_GNN, self).__init__()
        if gnn is None:
            return
        self.gnn = gnn
        self.params = nn.ModuleList()
        self.neg_queue_size = neg_queue_size
        self.link_dec_dict = {}
        self.neg_queue = {}

        for source_type in rem_edge_list:
            self.link_dec_dict[source_type] = {}
            self.neg_queue[source_type] = {}
            for relation_type in rem_edge_list[source_type]:
                print(source_type, relation_type)
                matcher = Matcher(gnn.n_hid, gnn.n_hid)
                self.neg_queue[source_type][relation_type] = torch.FloatTensor([]).to(device)
                self.link_dec_dict[source_type][relation_type] = matcher
                self.params.append(matcher)
        self.attr_decoder = attr_decoder
        self.init_emb = nn.Parameter(torch.randn(gnn.in_dim))
        self.ce = nn.CrossEntropyLoss(reduction = 'none')
        self.neg_samp_num = neg_samp_num 
        
    def neg_sample(self, souce_node_list, pos_node_list):
        np.random.shuffle(souce_node_list)
        neg_nodes = []
        keys = {key : True for key in pos_node_list} 
        tot = 0
        for node_id in souce_node_list:
            if node_id not in keys:
                neg_nodes += [node_id]
                tot += 1
            if tot == self.neg_samp_num:
                break
        return neg_nodes
        
    def forward(self, node_feature, node_type, edge_time, edge_index, edge_type):
        return self.gnn(node_feature, node_type, edge_time, edge_index, edge_type)
    def link_loss(self, node_emb, rem_edge_list, ori_edge_list, node_dict, target_type, use_queue = True, update_queue = False):
        losses = 0
        ress   = []
        for source_type in rem_edge_list:
            if source_type not in self.link_dec_dict:
                continue
            for relation_type in rem_edge_list[source_type]:
                if relation_type not in self.link_dec_dict[source_type]:
                    continue
                rem_edges = rem_edge_list[source_type][relation_type]
                if len(rem_edges) <= 8:
                    continue
                ori_edges = ori_edge_list[source_type][relation_type]
                matcher = self.link_dec_dict[source_type][relation_type]

                target_ids, positive_source_ids = rem_edges[:,0].reshape(-1, 1), rem_edges[:,1].reshape(-1, 1)
                n_nodes = len(target_ids)
                source_node_ids = np.unique(ori_edges[:, 1])
                
                
                negative_source_ids = [self.neg_sample(source_node_ids, ori_edges[ori_edges[:, 0] == t_id][:, 1].tolist()) for t_id in target_ids]
                sn = min([len(neg_ids) for neg_ids in negative_source_ids])  # 统计所有样本点的负样本list的最短长度

                negative_source_ids = [neg_ids[:sn] for neg_ids in negative_source_ids]  # 统一选择最短长度

                
                source_ids = torch.LongTensor(np.concatenate((positive_source_ids, negative_source_ids), axis=-1) + node_dict[source_type][0])
                emb = node_emb[source_ids]
                
                
                if use_queue and len(self.neg_queue[source_type][relation_type]) // n_nodes > 0:
                    tmp = self.neg_queue[source_type][relation_type]
                    stx = len(tmp) // n_nodes
                    tmp = tmp[: stx * n_nodes].reshape(n_nodes, stx, -1)
                    rep_size = sn + 1 + stx
                    source_emb = torch.cat([emb, tmp], dim=1)
                    source_emb = source_emb.reshape(n_nodes * rep_size, -1)
                else:
                    rep_size = sn + 1
                    source_emb = emb.reshape(source_ids.shape[0] * rep_size, -1)
                    
                target_ids = target_ids.repeat(rep_size, 1) + node_dict[target_type][0]
                target_emb = node_emb[target_ids.reshape(-1)]
                res = matcher.forward(target_emb, source_emb)
                res = res.reshape(n_nodes, rep_size)
                ress += [res.detach()]
                losses += F.log_softmax(res, dim=-1)[:,0].mean()
                if update_queue and 'L1' not in relation_type and 'L2' not in relation_type:
                    tmp = self.neg_queue[source_type][relation_type]
                    self.neg_queue[source_type][relation_type] = \
                        torch.cat([node_emb[source_node_ids].detach(), tmp], dim=0)[:int(self.neg_queue_size * n_nodes)]
        return -losses / len(ress), ress

    
    def text_loss(self, reps, texts, w2v_model, device):
        def parse_text(texts, w2v_model, device):
            idxs = []
            pad  = w2v_model.wv.key_to_index['eos']
            # text to tokens
            for text in texts:
                idx = []
                for word in ['bos'] + preprocess_string(text) + ['eos']:
                    if word in w2v_model.wv.key_to_index:
                        idx += [w2v_model.wv.key_to_index[word]]
                idxs += [idx]
            mxl = np.max([len(s) for s in idxs]) + 1
            inp_idxs = []
            out_idxs = []
            masks    = []
            for i, idx in enumerate(idxs):
                inp_idxs += [idx + [pad for _ in range(mxl - len(idx) - 1)]]
                out_idxs += [idx[1:] + [pad for _ in range(mxl - len(idx))]] 
                masks    += [[1 for _ in range(len(idx))] + [0 for _ in range(mxl - len(idx) - 1)]]
            return torch.LongTensor(inp_idxs).transpose(0, 1).to(device), \
                   torch.LongTensor(out_idxs).transpose(0, 1).to(device), torch.BoolTensor(masks).transpose(0, 1).to(device)
        inp_idxs, out_idxs, masks = parse_text(texts, w2v_model, device)
        pred_prob = self.attr_decoder(inp_idxs, reps.repeat(inp_idxs.shape[0], 1, 1))      
        return self.ce(pred_prob[masks], out_idxs[masks]).mean()

    def feat_loss(self, reps, out):
        return -self.attr_decoder(reps, out).mean()


class Classifier(nn.Module):
    r"""Classifier for graph node classification task.

    Parameters
    ----------
    n_hid: int
        Input size.
    n_out: int
        Output size.

    """
    def __init__(self, n_hid, n_out):
        super(Classifier, self).__init__()
        self.n_hid    = n_hid
        self.n_out    = n_out
        self.linear   = nn.Linear(n_hid,  n_out)
    def forward(self, x):
        tx = self.linear(x)
        return torch.log_softmax(tx.squeeze(), dim=-1)
    def __repr__(self):
        return '{}(n_hid={}, n_out={})'.format(
            self.__class__.__name__, self.n_hid, self.n_out)

    
class Matcher(nn.Module):
    r"""Matching between a pair of nodes to conduct link prediction.
        Use multi-head attention as matching model.

    Parameters
    ----------
    n_hid: int
        Input size.
    n_out: int
        Output size.
    temperature: float, optional
        Temperature.
        (default: :obj:`0.1`)

    """
    
    def __init__(self, n_hid, n_out, temperature=0.1):
        super(Matcher, self).__init__()
        self.n_hid       = n_hid
        self.linear      = nn.Linear(n_hid,  n_out)
        self.sqrt_hd     = math.sqrt(n_out)
        self.drop        = nn.Dropout(0.2)
        self.cosine      = nn.CosineSimilarity(dim=1)
        self.cache       = None
        self.temperature = temperature
    def forward(self, x, ty, use_norm = True):
        tx = self.drop(self.linear(x))
        if use_norm:
            return self.cosine(tx, ty) / self.temperature
        else:
            return (tx * ty).sum(dim=-1) / self.sqrt_hd
    def __repr__(self):
        return '{}(n_hid={})'.format(
            self.__class__.__name__, self.n_hid)

    
class HGT(nn.Module):
    r"""The Heterogeneous Graph Transformer (HGT) operator from the
    `"Heterogeneous Graph Transformer" 
    <https://arxiv.org/abs/2003.01332>`_ paper.
    """
    
    def __init__(self, in_dim, n_hid, num_types, num_relations, n_heads, n_layers, dropout = 0.2, prev_norm = False, last_norm = False):
        super(HGT, self).__init__()
        self.gcs = nn.ModuleList()
        self.num_types = num_types
        self.in_dim    = in_dim
        self.n_hid     = n_hid
        self.adapt_ws  = nn.ModuleList()
        self.drop      = nn.Dropout(dropout)
        for t in range(num_types):
            self.adapt_ws.append(nn.Linear(in_dim, n_hid))
        for l in range(n_layers - 1):
            self.gcs.append(HGTConv(n_hid, n_hid, num_types, num_relations, n_heads, dropout, use_norm = prev_norm))
        self.gcs.append(HGTConv(n_hid, n_hid, num_types, num_relations, n_heads, dropout, use_norm = last_norm))

    def forward(self, node_feature, node_type, edge_time, edge_index, edge_type):
        res = torch.zeros(node_feature.size(0), self.n_hid).to(node_feature.device)
        for t_id in range(self.num_types):
            idx = (node_type == int(t_id))
            if idx.sum() == 0:
                continue
            res[idx] = torch.tanh(self.adapt_ws[t_id](node_feature[idx]))
        meta_xs = self.drop(res)
        del res
        for gc in self.gcs:
            meta_xs = gc(meta_xs, node_type, edge_index, edge_type, edge_time)
        return meta_xs   

    
class RNNModel(nn.Module):
    r"""Container module with an encoder, a recurrent module, and a decoder.

    Parameters
    ----------
    n_word: int
        Number of tokens.
    ninp: int
        Input size.
    nhid: int
        Hidden size.
    nlayers: int
        Layer number of LSTM.
    dropout: float, optional
        Dropout rate.
        (default: :obj:`0.2`)

    """

    def __init__(self, n_word, ninp, nhid, nlayers, dropout=0.2):
        super(RNNModel, self).__init__()
        self.drop = nn.Dropout(dropout)
        self.rnn = nn.LSTM(nhid, nhid, nlayers)
        self.encoder = nn.Embedding(n_word, nhid)
        self.decoder = nn.Linear(nhid, n_word)
        self.adp     = nn.Linear(ninp + nhid, nhid)
    def forward(self, inp, hidden = None):
        emb = self.encoder(inp)
        if hidden is not None:
            emb = torch.cat((emb, hidden), dim=-1)
            emb = F.gelu(self.adp(emb))
        output, _ = self.rnn(emb)
        decoded = self.decoder(self.drop(output))
        return decoded
    def from_w2v(self, w2v):
        initrange = 0.1
        self.encoder.weight.data = w2v
        self.decoder.weight = self.encoder.weight
        
        self.encoder.weight.requires_grad = False
        self.decoder.weight.requires_grad = False
        


import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.utils import softmax
from torch_geometric.nn.inits import glorot
from torch_geometric.nn.conv import MessagePassing


class HGTConv(MessagePassing):

    r"""The Heterogeneous Graph Transformer (HGT) operator from the
    `"Heterogeneous Graph Transformer" 
    <https://arxiv.org/abs/2003.01332>`_ paper.

    Parameters
    ----------
    in_dim: int
        Size of each input sample of every
        node type, or :obj:`-1` to derive the size from the first input(s)
        to the forward method.
    out_dim: int
        Size of each output sample.
    num_type: int
        Number of node types.
    num_relations: int
        Number of relations.
    heads: int, optional
        Number of multi-head-attentions.
        (default: :obj:`1`)
    dropout: float
        Dropout rate.
        (default: :obj:`0.2`)
    use_norm: bool, optional
        If use norm.
        (default: :obj:`True`)
    **kwargs: optional
        Additional arguments of
        :class:`torch_geometric.nn.conv.MessagePassing`.

    """

    def __init__(self, in_dim, out_dim, num_types, num_relations, n_heads=1, dropout=0.2, use_norm=True, **kwargs):
        super(HGTConv, self).__init__(node_dim=0, aggr='add', **kwargs)

        self.in_dim        = in_dim
        self.out_dim       = out_dim
        self.num_types     = num_types
        self.num_relations = num_relations
        self.total_rel     = num_types * num_relations * num_types
        self.n_heads       = n_heads
        self.d_k           = out_dim // n_heads
        self.sqrt_dk       = math.sqrt(self.d_k)
        self.use_norm      = use_norm
        self.att           = None
        
        
        self.k_linears   = nn.ModuleList()
        self.q_linears   = nn.ModuleList()
        self.v_linears   = nn.ModuleList()
        self.a_linears   = nn.ModuleList()
        self.norms       = nn.ModuleList()
        
        for t in range(num_types):
            self.k_linears.append(nn.Linear(in_dim,   out_dim))
            self.q_linears.append(nn.Linear(in_dim,   out_dim))
            self.v_linears.append(nn.Linear(in_dim,   out_dim))
            self.a_linears.append(nn.Linear(out_dim,  out_dim))
            if use_norm:
                self.norms.append(nn.LayerNorm(out_dim))
        '''
            TODO: make relation_pri smaller, as not all <st, rt, tt> pair exist in meta relation list.
        '''
        self.relation_pri   = nn.Parameter(torch.ones(num_relations, self.n_heads))
        self.relation_att   = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.relation_msg   = nn.Parameter(torch.Tensor(num_relations, n_heads, self.d_k, self.d_k))
        self.skip           = nn.Parameter(torch.ones(num_types))
        self.drop           = nn.Dropout(dropout)
        self.emb            = RelTemporalEncoding(in_dim)
        
        glorot(self.relation_att)
        glorot(self.relation_msg)
        
    def forward(self, node_inp, node_type, edge_index, edge_type, edge_time):
        return self.propagate(edge_index, node_inp=node_inp, node_type=node_type, \
                              edge_type=edge_type, edge_time = edge_time)

    def message(self, edge_index_i, node_inp_i, node_inp_j, node_type_i, node_type_j, edge_type, edge_time):
        
        # j: source, i: target; <j, i>
        data_size = edge_index_i.size(0)
        
        # Create Attention and Message tensor beforehand.
        res_att     = torch.zeros(data_size, self.n_heads).to(node_inp_i.device)
        res_msg     = torch.zeros(data_size, self.n_heads, self.d_k).to(node_inp_i.device)
        
        for source_type in range(self.num_types):
            sb = (node_type_j == int(source_type))
            k_linear = self.k_linears[source_type]
            v_linear = self.v_linears[source_type] 
            for target_type in range(self.num_types):
                tb = (node_type_i == int(target_type)) & sb
                q_linear = self.q_linears[target_type]
                for relation_type in range(self.num_relations):
                    
                    # idx is all the edges with meta relation <source_type, relation_type, target_type>
                    idx = (edge_type == int(relation_type)) & tb
                    if idx.sum() == 0:
                        continue
                    '''
                        Get the corresponding input node representations by idx.
                        Add tempotal encoding to source representation (j)
                    '''
                    target_node_vec = node_inp_i[idx]
                    source_node_vec = self.emb(node_inp_j[idx], edge_time[idx])

                    '''
                        Step 1: Heterogeneous Mutual Attention
                    '''
                    q_mat = q_linear(target_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = k_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    k_mat = torch.bmm(k_mat.transpose(1,0), self.relation_att[relation_type]).transpose(1,0)
                    res_att[idx] = (q_mat * k_mat).sum(dim=-1) * self.relation_pri[relation_type] / self.sqrt_dk
                    '''
                        Step 2: Heterogeneous Message Passing
                    '''
                    v_mat = v_linear(source_node_vec).view(-1, self.n_heads, self.d_k)
                    res_msg[idx] = torch.bmm(v_mat.transpose(1,0), self.relation_msg[relation_type]).transpose(1,0)   
        # Softmax based on target node's id (edge_index_i). Store attention value in self.att for later visualization.
        self.att = softmax(res_att, edge_index_i)
        res = res_msg * self.att.view(-1, self.n_heads, 1)
        del res_att, res_msg
        return res.view(-1, self.out_dim)


    def update(self, aggr_out, node_inp, node_type):

        aggr_out = F.gelu(aggr_out)
        res = torch.zeros(aggr_out.size(0), self.out_dim).to(node_inp.device)
        for target_type in range(self.num_types):
            idx = (node_type == int(target_type))
            if idx.sum() == 0:
                continue
            trans_out = self.a_linears[target_type](aggr_out[idx])
            '''
                Add skip connection with learnable weight self.skip[t_id]
            '''
            alpha = torch.sigmoid(self.skip[target_type])
            if self.use_norm:
                res[idx] = self.norms[target_type](trans_out * alpha + node_inp[idx] * (1 - alpha))
            else:
                res[idx] = trans_out * alpha + node_inp[idx] * (1 - alpha)
        return self.drop(res)

    def __repr__(self):
        return '{}(in_dim={}, out_dim={}, num_types={}, num_types={})'.format(
            self.__class__.__name__, self.in_dim, self.out_dim,
            self.num_types, self.num_relations)


class RelTemporalEncoding(nn.Module):
    '''
        Implement the Temporal Encoding (Sinusoid) function.
    '''
    def __init__(self, n_hid, max_len = 240, dropout = 0.2):
        super(RelTemporalEncoding, self).__init__()
        self.drop = nn.Dropout(dropout)
        position = torch.arange(0., max_len).unsqueeze(1)
        div_term = 1 / (10000 ** (torch.arange(0., n_hid * 2, 2.)) / n_hid / 2)
        self.emb = nn.Embedding(max_len, n_hid * 2)
        self.emb.weight.data[:, 0::2] = torch.sin(position * div_term) / math.sqrt(n_hid)
        self.emb.weight.data[:, 1::2] = torch.cos(position * div_term) / math.sqrt(n_hid)
        self.emb.requires_grad = False
        self.lin = nn.Linear(n_hid * 2, n_hid)
    def forward(self, x, t):
        return x + self.lin(self.drop(self.emb(t)))


import dgl
import torch
import random
from .utils import save_pkl_file, open_pkl_file
from dgl.data.utils import save_graphs, load_graphs


def construct_graph(data_dir, graph, src_dst2edge_type):
    r"""

    Construct dgl.heterograph from ggfm.data.graph.

    Parameters
    ----------
    data_dir: str
        Data directory for saving dgl.heterograph object, which is saved as data_dir/graph.bin.

    graph: class:`ggfm.data.Graph`
        Target graph.
    
    src_dst2edge_type: dict
        The edge types corresponding to (src, dst) types.

    """

    edges = graph.edge_list

    single_edges = {}
    
    for target_type in edges:
            for source_type in edges[target_type]:
                for relation_type in edges[target_type][source_type]:
                    srcs, dsts = [], []
                    single_edges[(source_type, target_type, relation_type)] = [[], []]
                    for target_id in edges[target_type][source_type][relation_type]:
                        for source_id in edges[target_type][source_type][relation_type][target_id]:
                            srcs.append(source_id)
                            dsts.append(target_id)
                    single_edges[(source_type, target_type, relation_type)][0].extend(srcs)
                    single_edges[(source_type, target_type, relation_type)][1].extend(dsts)
    
    merged_edges = {}
    for key, values in single_edges.items():
        src_type, dst_type = key[0], key[1]
        srcs, dsts = values[0], values[1]
        if src_type == "paper" and dst_type == "paper":
            cur_relation = key[2]
            if "rev" in cur_relation:
                cur_edge_type = (src_type, "cited by", dst_type)
                if cur_edge_type not in merged_edges:
                    merged_edges[cur_edge_type] = [[], []]
                merged_edges[cur_edge_type][0].extend(srcs)
                merged_edges[cur_edge_type][1].extend(dsts)
            else:
                cur_edge_type = (src_type, "cites", dst_type)
                if cur_edge_type not in merged_edges:
                    merged_edges[cur_edge_type] = [[], []]
                merged_edges[cur_edge_type][0].extend(srcs)
                merged_edges[cur_edge_type][1].extend(dsts)
        elif src_type == "field" and dst_type == "field":
            cur_relation = key[2]
            if "rev" in cur_relation:  # contains
                cur_edge_type = (src_type, "contains", dst_type)
                if cur_edge_type not in merged_edges:
                    merged_edges[cur_edge_type] = [[], []]
                merged_edges[cur_edge_type][0].extend(srcs)
                merged_edges[cur_edge_type][1].extend(dsts)
            else:
                cur_edge_type = (src_type, "in", dst_type)
                if cur_edge_type not in merged_edges:
                    merged_edges[cur_edge_type] = [[], []]
                merged_edges[cur_edge_type][0].extend(srcs)
                merged_edges[cur_edge_type][1].extend(dsts)
        else:
            cur_edge_type = (src_type, src_dst2edge_type[(src_type, dst_type)], dst_type)
            if cur_edge_type not in merged_edges:
                merged_edges[cur_edge_type] = [[], []]
            merged_edges[cur_edge_type][0].extend(srcs)
            merged_edges[cur_edge_type][1].extend(dsts)
    
    for key, values in merged_edges.items():
        merged_edges[key] = (torch.tensor(values[0]), torch.tensor(values[1]))

    g = dgl.heterograph(merged_edges)
    save_graphs(data_dir + "graph.bin", g)
    print("graph has been saved!")


def construct_graph_node_name(data_dir, graph):
    r"""

    Construct graph_node_name.pkl for ggfm.data.graph.

    Parameters
    ----------
    data_dir: str
        Data directory for saving dgl.heterograph object, which is saved as data_dir/graph_node_name.pkl.

    graph: class:`ggfm.data.Graph`
        Target graph.

    """
    graph_node_name = {}
    graph_node_type = graph.get_types()
    for i in range(len(graph_node_type)):
        attr = "name"
        if graph_node_type[i] == "paper": attr = "title"
        graph_node_name[graph_node_type[i]] = graph.node_feature[graph_node_type[i]][attr].tolist()
    save_pkl_file(data_dir + "graph_node_name.pkl", graph_node_name)


def metapath_based_corpus_construction(data_dir, target_type, metapaths, relation, mid_types, labeled_node_idxs, k=2):

    r"""

    Metapath-based corpus construction.

    Parameters
    ----------
    data_dir: str
        Data directory for loading graph.bin and graph_node_name.pkl.
    
    target_type: str
        Sampled node type.
    
    metapaths: list
        Sampled metapaths for each node type.
    
    relation: list
        Relations for each node type.
    
    mid_types: list
        Midtypes of each node type's metapaths.
    
    labeled_node_idxs: list
        Sampled node indexes for the target node type.
    
    k: int, optional
        Number of samples retained for each metapath sampling。
        (default: :obj:`2`)

    """

    glist, label_dict = load_graphs(data_dir + "graph.bin")
    g = glist[0]

    graph_node_name = open_pkl_file(data_dir + 'graph_node_name.pkl')
    metapath = metapaths[target_type]
    relation = relation[target_type]

    sampling_time = 10
    num_nodes = len(labeled_node_idxs)
        
    all_path_for_sampling_times = [[] for _ in range(num_nodes)]
    print("---------------------------------------")
    print(f"Sampling {target_type} type nodes...")

    for p, path in enumerate(metapath):
        path_for_sampling_times = [[] for _ in range(num_nodes)]
        print(f"Sampling the {p}-th path...")
        for st in range(sampling_time):
            traces, types = dgl.sampling.random_walk(g=g, nodes=torch.tensor(labeled_node_idxs), metapath=path)
            traces = traces.tolist()
            length = len(traces)
            print(f"Performing the {st}-th sampling...")
            for i in range(length):
                if i % 10000 == 0:
                    print(f"Sampled {i} nodes...")
                path_i = []
                if traces[i][1] != -1: 
                    path_i.append(target_type.replace('_', ' '))
                    path_i.append(graph_node_name[target_type][traces[i][0]].replace('_', ' '))
                    path_i.append(relation[p][0])
                    mid_type = mid_types[target_type][p][0]
                    path_i.append(mid_type.replace('_', ' '))
                    path_i.append(graph_node_name[mid_type][traces[i][1]].replace('_', ' '))
                    path_i.append(relation[p][1])
                    if len(traces[i]) <= 3:  # apa pa
                        path_i.append(target_type.replace('_', ' '))
                        path_i.append(graph_node_name[target_type][traces[i][2]].replace('_', ' '))
                    else:  # apcpa pcpa
                        mid_type = mid_types[target_type][p][1]
                        path_i.append(mid_type.replace('_', ' '))
                        path_i.append(graph_node_name[mid_type][traces[i][2]].replace('_', ' '))
                        path_i.append(relation[p][2])

                        mid_type = mid_types[target_type][p][2]
                        path_i.append(mid_type.replace('_', ' '))
                        path_i.append(graph_node_name[mid_type][traces[i][3]].replace('_', ' '))
                        path_i.append(relation[p][3])

                        mid_type = target_type
                        path_i.append(mid_type.replace('_', ' '))
                        path_i.append(graph_node_name[mid_type][traces[i][4]].replace('_', ' '))
                else:
                    path_i.append(target_type.replace('_', ' '))
                    path_i.append(graph_node_name[target_type][traces[i][0]].replace('_', ' '))
                
                path_i.append("</s>")
                path_i = " ".join(path_i)
                path_for_sampling_times[i].append(path_i) 
        path_for_sampling_times = [list(set(item)) for item in path_for_sampling_times if item]
        
        path_for_sampling_times = [random.sample(item, min(k, len(item))) for item in path_for_sampling_times if item]
        for i, item in enumerate(path_for_sampling_times):
            path_for_sampling_times[i] = " ".join(item)
            all_path_for_sampling_times[i].append(path_for_sampling_times[i])
    all_path_for_sampling_times = [list(set(item)) for item in all_path_for_sampling_times if item]
    for i, item in enumerate(all_path_for_sampling_times):
        all_path_for_sampling_times[i] = " ".join(item)
        all_path_for_sampling_times[i].rstrip(' </s> ')

    print(all_path_for_sampling_times[0])
    print(f"length: {len(all_path_for_sampling_times)}")
    return all_path_for_sampling_times


import contextlib
import torch
import torch.nn as nn
from torch.cuda.amp import autocast as autocast
from transformers import AutoModelForCausalLM, AutoTokenizer
from torch_scatter import scatter

BOS = '<s>[INST]'
EOS_USER = '[/INST]'
EOS = '[/s]'

IGNORE_INDEX = -100


class LLAGA(torch.nn.Module):
    r"""`"LLaGA: Large Language and Graph Assistant"
        <https://arxiv.org/abs/2402.08170>`_ paper.

        Parameters
        ----------
        model: class:`ggfm.models`
            The used model.
        tokenizer: `transformers.AutoTokenizer`
            The tokenizer.
        device: int
            Device

        """
    def __init__(
        self,
        args,
        **kwargs
    ):
        super().__init__()
        self.max_txt_len = args.max_txt_len
        self.max_new_tokens = args.max_new_tokens

        print('Loading LLAMA')
        kwargs = {
            "max_memory": {1: '60GiB'},
            "device_map": "auto",
            "revision": "main",
        }

        self.tokenizer = AutoTokenizer.from_pretrained(args.llm_model_path, use_fast=False, revision=kwargs["revision"])
        self.tokenizer.pad_token_id = 0
        self.tokenizer.padding_side = 'left'

        model = AutoModelForCausalLM.from_pretrained(
            args.llm_model_path,
            torch_dtype=torch.float16,
            low_cpu_mem_usage=True,
            **kwargs
        )

        if args.llm_frozen == True:
            print("Freezing LLAMA!")
            for name, param in model.named_parameters():
                param.requires_grad = False
        else:
            raise ValueError

        self.model = model
        print('Finish loading LLAMA!')

        self.projector = nn.Sequential(
            nn.Linear(args.mm_hidden_size, 2048),
            nn.Linear(2048, 4096),
        ).to(self.model.device)

        self.word_embedding = self.model.model.get_input_embeddings()
    @property
    def device(self):
        return list(self.parameters())[0].device

    def maybe_autocast(self, dtype=torch.bfloat16):
        # if on cpu, don't use autocast
        # if on gpu, use autocast with dtype if provided, otherwise use torch.float16
        enable_autocast = self.device != torch.device("cpu")

        if enable_autocast:
            return torch.cuda.amp.autocast(dtype=dtype)
        else:
            return contextlib.nullcontext()

    def encode_graphs(self, samples):
        graphs = samples['graph']
        graphs = graphs.to(self.model.device)
        n_embeds, _ = self.graph_encoder(graphs.x, graphs.edge_index.long(), graphs.edge_attr)

        # mean pooling
        g_embeds = scatter(n_embeds, graphs.batch, dim=0, reduce='mean')

        return g_embeds

    def forward(self, samples):

        # encode description, questions and labels
        questions = self.tokenizer(samples["question"], add_special_tokens=False)
        label = self.tokenizer(samples["label"], add_special_tokens=False)

        graph_id = samples["graph_id"]
        graph = samples["label"]
        graph_emb = samples["graph_emb"]

        # encode special tokens
        eos_tokens = self.tokenizer(EOS, add_special_tokens=False)
        eos_user_tokens = self.tokenizer(EOS_USER, add_special_tokens=False)
        bos_embeds = self.word_embedding(self.tokenizer(BOS, add_special_tokens=False, return_tensors='pt').input_ids[0].to(self.device))
        pad_embeds = self.word_embedding(torch.tensor(self.tokenizer.pad_token_id).to(self.device)).unsqueeze(0)

        graph_emb = torch.nn.utils.rnn.pad_sequence(graph_emb, batch_first=True, padding_value=0)
        graph_embeds = self.projector(graph_emb.to(self.device))

        batch_size = len(samples['id'])
        batch_inputs_embeds = []
        batch_attention_mask = []
        batch_label_input_ids = []
        for i in range(batch_size):
            # Add bos & eos token
            label_input_ids = label.input_ids[i][:self.max_new_tokens] + eos_tokens.input_ids
            input_ids = questions.input_ids[i] + eos_user_tokens.input_ids + label_input_ids
            inputs_embeds = self.word_embedding(torch.tensor(input_ids).to(self.model.device))
            inputs_embeds = torch.cat([bos_embeds, graph_embeds[i], inputs_embeds], dim=0)

            batch_inputs_embeds.append(inputs_embeds)
            batch_attention_mask.append([1] * inputs_embeds.shape[0])
            label_input_ids = [IGNORE_INDEX] * (inputs_embeds.shape[0]-len(label_input_ids))+label_input_ids
            batch_label_input_ids.append(label_input_ids)

        # pad inputs_embeds
        max_length = max([x.shape[0] for x in batch_inputs_embeds])
        for i in range(batch_size):
            pad_length = max_length-batch_inputs_embeds[i].shape[0]
            batch_inputs_embeds[i] = torch.cat([pad_embeds.repeat(pad_length, 1), batch_inputs_embeds[i]])
            batch_attention_mask[i] = [0]*pad_length+batch_attention_mask[i]
            batch_label_input_ids[i] = [IGNORE_INDEX] * pad_length+batch_label_input_ids[i]

        inputs_embeds = torch.stack(batch_inputs_embeds, dim=0).to(self.model.device)
        attention_mask = torch.tensor(batch_attention_mask).to(self.model.device)
        label_input_ids = torch.tensor(batch_label_input_ids).to(self.model.device)

        with self.maybe_autocast():
            outputs = self.model(
                inputs_embeds=inputs_embeds,
                attention_mask=attention_mask,
                return_dict=True,
                labels=label_input_ids,
            )

        return outputs.loss

    def inference(self, samples):

        # encode description, questions and labels
        questions = self.tokenizer(samples["question"], add_special_tokens=False)

        graph_id = samples["graph_id"]
        graph = samples["label"]
        graph_emb = samples["graph_emb"]

        # encode special tokens
        eos_tokens = self.tokenizer(EOS, add_special_tokens=False)
        eos_user_tokens = self.tokenizer(EOS_USER, add_special_tokens=False)
        bos_embeds = self.word_embedding(
            self.tokenizer(BOS, add_special_tokens=False, return_tensors='pt').input_ids[0].to(self.device))
        pad_embeds = self.word_embedding(torch.tensor(self.tokenizer.pad_token_id).to(self.device)).unsqueeze(0)

        graph_emb = torch.nn.utils.rnn.pad_sequence(graph_emb, batch_first=True, padding_value=0)
        graph_embeds = self.projector(graph_emb.to(self.device))

        batch_size = len(samples['id'])
        batch_inputs_embeds = []
        batch_attention_mask = []
        for i in range(batch_size):
            # Add bos & eos token
            input_ids = questions.input_ids[i] + eos_user_tokens.input_ids
            inputs_embeds = self.word_embedding(torch.tensor(input_ids).to(self.model.device))
            inputs_embeds = torch.cat([bos_embeds, graph_embeds[i], inputs_embeds], dim=0)
            batch_inputs_embeds.append(inputs_embeds)
            batch_attention_mask.append([1] * inputs_embeds.shape[0])

        # pad inputs_embeds
        max_length = max([x.shape[0] for x in batch_inputs_embeds])
        for i in range(batch_size):
            pad_length = max_length - batch_inputs_embeds[i].shape[0]
            batch_inputs_embeds[i] = torch.cat([pad_embeds.repeat(pad_length, 1), batch_inputs_embeds[i]])
            batch_attention_mask[i] = [0] * pad_length + batch_attention_mask[i]

        inputs_embeds = torch.stack(batch_inputs_embeds, dim=0).to(self.model.device)
        attention_mask = torch.tensor(batch_attention_mask).to(self.model.device)

        with self.maybe_autocast():
            outputs = self.model.generate(
                inputs_embeds=inputs_embeds,
                max_new_tokens=self.max_new_tokens,
                attention_mask=attention_mask,
                use_cache=True
            )

        pred = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

        return {'id': samples['id'],
                'pred': pred,
                'label': samples['label'],
                'question': samples['question'],
                }



    def print_trainable_params(self):
        trainable_params = 0
        all_param = 0

        for _, param in self.named_parameters():
            num_params = param.numel()

            all_param += num_params
            if param.requires_grad:
                trainable_params += num_params

        return trainable_params, all_param



from .graph import Graph, RenameUnpickler, renamed_load
from .hgsampling import sample_subgraph, feature_extractor
from .random_walk import construct_link_and_node, random_walk_based_corpus_construction, get_type_id
from .utils import open_pkl_file, open_txt_file, save_pkl_file, save_txt_file, mean_reciprocal_rank, ndcg_at_k, args_print
from .lm_generate_embs import generate_lm_embs
from .metapath import construct_graph, construct_graph_node_name, metapath_based_corpus_construction

__all__ = [
    'Graph',
    'sample_subgraph',
    'open_pkl_file',
    'open_txt_file',
    'save_pkl_file',
    'save_txt_file',
    'renamed_load',
    'mean_reciprocal_rank',
    'ndcg_at_k',
    'feature_extractor',
    'args_print',
    'construct_link_and_node',
    'random_walk_based_corpus_construction',
    'get_type_id',
    'generate_lm_embs',
    'construct_graph',
    'construct_graph_node_name',
    'metapath_based_corpus_construction'
]

classes = __all__



import torch
import argparse
import numpy as np
from texttable import Texttable
from collections import defaultdict
from ggfm.models import HGT, RNNModel, Matcher, GPT_GNN, get_optimizer
from ggfm.data import renamed_load, sample_subgraph, ndcg_at_k, feature_extractor
from warnings import filterwarnings
filterwarnings("ignore")


def args_print(args):
    _dict = vars(args)
    t = Texttable() 
    t.add_row(["Parameter", "Value"])
    for k in _dict:
        t.add_row([k, _dict[k]])
    print(t.draw())


def randint():
    return np.random.randint(2**32 - 1)


def to_torch(feature, time, edge_list, graph):
    '''
        Transform a sampled sub-graph into pytorch Tensor
        node_dict: {node_type: <node_number, node_type_ID>} node_number is used to trace back the nodes in original graph.
        edge_dict: {edge_type: edge_type_ID}
    '''
    node_dict = {}
    node_feature = []
    node_type    = []
    node_time    = []
    edge_index   = []
    edge_type    = []
    edge_time    = []
    
    node_num = 0
    types = graph.get_types()
    for t in types:
        node_dict[t] = [node_num, len(node_dict)]
        node_num += len(feature[t])
        
    for t in types:
        node_feature += list(feature[t])
        node_time    += list(time[t])
        node_type    += [node_dict[t][1] for _ in range(len(feature[t]))]
        
    edge_dict = {e[2]: i for i, e in enumerate(graph.get_meta_graph())}
    edge_dict['self'] = len(edge_dict)

    # transfer to homogeneous graph
    for target_type in edge_list:
        for source_type in edge_list[target_type]:
            for relation_type in edge_list[target_type][source_type]:
                for ii, (ti, si) in enumerate(edge_list[target_type][source_type][relation_type]):
                    tid, sid = ti + node_dict[target_type][0], si + node_dict[source_type][0]
                    edge_index += [[sid, tid]]
                    edge_type  += [edge_dict[relation_type]]

                    # time ranges from 1900 - 2020, largest span is 120.
                    edge_time  += [node_time[tid] - node_time[sid] + 120]

    node_feature = torch.FloatTensor(node_feature)
    node_type    = torch.LongTensor(node_type)
    edge_time    = torch.LongTensor(edge_time)
    edge_index   = torch.LongTensor(edge_index).t()
    edge_type    = torch.LongTensor(edge_type)
    return node_feature, node_type, edge_time, edge_index, edge_type, node_dict, edge_dict


def node_sample(seed, graph, target_nodes, time_range, batch_size):
    np.random.seed(seed)
    samp_target_nodes = target_nodes[np.random.choice(len(target_nodes), batch_size)]
    threshold = 0.5
    
    feature, times, edge_list, _, attr = sample_subgraph(graph, time_range, inp = {target_type: samp_target_nodes}, \
                                                         feature_extractor = feature_extractor, sampled_depth = args.sample_depth, \
                                                         sampled_number = args.sample_width)
    

    rem_edge_list = defaultdict(  # source_type
                        lambda: defaultdict(  # relation_type
                            lambda: []  # [target_id, source_id] 
                                ))
    
    # source_type, relation_type, edges[]
    ori_list = {} 
    for source_type in edge_list[target_type]:
        ori_list[source_type] = {}
        for relation_type in edge_list[target_type][source_type]:
            ori_list[source_type][relation_type] = np.array(edge_list[target_type][source_type][relation_type])
            el = []
            for target_ser, source_ser in edge_list[target_type][source_type][relation_type]:
                if relation_type not in rel_stop_list and target_ser < batch_size and np.random.random() > threshold:
                    rem_edge_list[source_type][relation_type] += [[target_ser, source_ser]]
                    continue
                el += [[target_ser, source_ser]]
            el = np.array(el)
            edge_list[target_type][source_type][relation_type] = el
            
            if relation_type == 'self':
                continue
            else:
                if 'rev_' in relation_type:
                    rev_relation = relation_type[4:]
                else:
                    rev_relation = 'rev_' + relation_type
                edge_list[source_type]['paper'][rev_relation] = list(np.stack((el[:,1], el[:,0])).T)
                
    '''
        Adding feature nodes:
    '''
    n_target_nodes = len(feature[target_type])
    feature[target_type] = np.concatenate((feature[target_type], np.zeros([batch_size, feature[target_type].shape[1]])))
    times[target_type]   = np.concatenate((times[target_type], times[target_type][:batch_size]))

    for source_type in edge_list[target_type]:
        for relation_type in edge_list[target_type][source_type]:
            el = []
            for target_ser, source_ser in edge_list[target_type][source_type][relation_type]:
                if target_ser < batch_size:
                    if relation_type == 'self':
                        el += [[target_ser + n_target_nodes, target_ser + n_target_nodes]]
                    else:
                        el += [[target_ser + n_target_nodes, source_ser]]
            if len(el) > 0:
                edge_list[target_type][source_type][relation_type] = \
                    np.concatenate((edge_list[target_type][source_type][relation_type], el))


    rem_edge_lists = {}
    for source_type in rem_edge_list:
        rem_edge_lists[source_type] = {}
        for relation_type in rem_edge_list[source_type]:
            rem_edge_lists[source_type][relation_type] = np.array(rem_edge_list[source_type][relation_type])
    del rem_edge_list
    
    return to_torch(feature, times, edge_list, graph), rem_edge_lists, ori_list, \
            attr[:batch_size], (n_target_nodes, n_target_nodes + batch_size)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Pre-training GPT-GNN on a given graph (heterogeneous / homogeneous)')

    '''
    GPT-GNN arguments 
    '''
    parser.add_argument('--attr_ratio', type=float, default=0.5, help='Ratio of attr-loss against link-loss, range: [0-1]') 
    parser.add_argument('--attr_type', type=str, default='text', choices=['text', 'vec'], help='The type of attribute decoder')
    parser.add_argument('--neg_samp_num', type=int, default=255, help='Maximum number of negative sample for each target node.')
    parser.add_argument('--queue_size', type=int, default=256, help='Max size of adaptive embedding queue.')
    parser.add_argument('--w2v_dir', type=str, default='/home/yjy/heteroPrompt/ggfm/ggfm/pretrained_model/w2v_all', help='The address of w2v_model.')

    '''
        Dataset arguments
    '''
    parser.add_argument('--data_dir', type=str, default='/home/yjy/heteroPrompt/ggfm/ggfm/datasets/graph.pk', help='The address of data for pretrain.')
    parser.add_argument('--pretrained_model_dir', type=str, default='/home/yjy/heteroPrompt/ggfm/ggfm/pretrained_model/gta_all_cs3', help='The save address for the pretrained model.')
    parser.add_argument('--cuda', type=int, default=1, help='Avaiable GPU ID')      
    parser.add_argument('--sample_depth', type=int, default=6, help='How many layers within a mini-batch subgraph')
    parser.add_argument('--sample_width', type=int, default=128, help='How many nodes to be sampled per layer per type')

    '''
    Model arguments 
    '''
    parser.add_argument('--conv_name', type=str, default='hgt', help='The name of GNN filter. By default is Heterogeneous Graph Transformer (hgt)')
    parser.add_argument('--n_hid', type=int, default=400, help='Number of hidden dimension')
    parser.add_argument('--n_heads', type=int, default=8, help='Number of attention head')
    parser.add_argument('--n_layers', type=int, default=3, help='Number of GNN layers')
    parser.add_argument('--prev_norm', help='Whether to add layer-norm on the previous layers', action='store_true')
    parser.add_argument('--last_norm', help='Whether to add layer-norm on the last layers', action='store_true')
    parser.add_argument('--dropout', type=int, default=0.2, help='Dropout ratio')

    '''
        Optimization arguments
    '''
    parser.add_argument('--max_lr', type=float, default=1e-3, help='Maximum learning rate.')
    parser.add_argument('--scheduler', type=str, default='cycle', help='Name of learning rate scheduler.' , choices=['cycle', 'cosine'])
    parser.add_argument('--n_epoch', type=int, default=20, help='Number of epoch to run')
    parser.add_argument('--n_pool', type=int, default=8, help='Number of process to sample subgraph')    
    parser.add_argument('--n_batch', type=int, default=32, help='Number of batch (sampled graphs) for each epoch') 
    parser.add_argument('--batch_size', type=int, default=256, help='Number of output nodes for training')    
    parser.add_argument('--clip', type=float, default=0.5, help='Gradient Norm Clipping') 


    args = parser.parse_args()
    args_print(args)

    if args.cuda != -1: device = torch.device("cuda:" + str(args.cuda))
    else: device = torch.device("cpu")

    graph = renamed_load(open(args.data_dir, 'rb'))

    pre_range   = {t: True for t in graph.times if t != None and t < 2014}
    train_range = {t: True for t in graph.times if t != None and t >= 2014  and t <= 2016}

    # Pretraining at any cost
    # generative pretraining
    pre_target_nodes   = []
    train_target_nodes = []
    target_type = 'paper'
    rel_stop_list = ['self', 'rev_PF_in_L0', 'rev_PF_in_L5', 'rev_PV_Repository', 'rev_PV_Patent']

    for p_id, _time in graph.node_feature[target_type]['time'].items():
        if _time in pre_range:
            pre_target_nodes += [[p_id, _time]]
        elif _time in train_range:
            train_target_nodes += [[p_id, _time]]
    pre_target_nodes = np.array(pre_target_nodes)
    train_target_nodes = np.array(train_target_nodes)


    repeat_num = int(len(pre_target_nodes) / args.batch_size // args.n_batch)

    data, rem_edge_list, ori_edge_list, _, _ = node_sample(randint(), graph, pre_target_nodes, pre_range, args.batch_size)
    node_feature, node_type, edge_time, edge_index, edge_type, node_dict, edge_dict = data
    types = graph.get_types()

    gnn = HGT(conv_name = args.conv_name, in_dim = len(graph.node_feature[target_type]['emb'].values[0]) + 401, n_hid = args.n_hid, \
            n_heads = args.n_heads, n_layers = args.n_layers, dropout = args.dropout, num_types = len(types), \
            num_relations = len(graph.get_meta_graph()) + 1, prev_norm = args.prev_norm, last_norm = args.last_norm)


    if args.attr_type == 'text':  
        from gensim.models import Word2Vec
        w2v_model = Word2Vec.load(args.w2v_dir)
        n_tokens = len(w2v_model.wv.key_to_index)
        attr_decoder = RNNModel(n_word = n_tokens, ninp = gnn.n_hid, \
                nhid = w2v_model.vector_size, nlayers = 2)
        attr_decoder.from_w2v(torch.FloatTensor(w2v_model.wv.vectors))
    else:
        attr_decoder = Matcher(gnn.n_hid, gnn.in_dim)
        
    gpt_gnn = GPT_GNN(gnn = gnn, rem_edge_list = rem_edge_list, attr_decoder = attr_decoder, \
                    neg_queue_size = 0, types = types, neg_samp_num = args.neg_samp_num, device = device)
    gpt_gnn.init_emb.data = node_feature[node_type == node_dict[target_type][1]].mean(dim=0).detach()
    gpt_gnn = gpt_gnn.to(device)



    best_val   = 100000
    train_step = 0
    stats = []
    optimizer_args = dict(lr=args.max_lr, weight_decay=1e-2, eps=1e-06)
    optimizer = torch.optim.AdamW(gpt_gnn.parameters(), **optimizer_args)

    if args.scheduler == 'cycle':
        scheduler = torch.optim.lr_scheduler.OneCycleLR(optimizer, pct_start=0.02, anneal_strategy='linear', final_div_factor=100,\
                            max_lr = args.max_lr, total_steps = repeat_num * args.n_batch * args.n_epoch + 1)
    elif args.scheduler == 'cosine':
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, repeat_num * args.n_batch, eta_min=1e-6)

    for epoch in np.arange(args.n_epoch) + 1:
        gpt_gnn.neg_queue_size = args.queue_size * epoch // args.n_epoch
        for batch in np.arange(repeat_num) + 1:
            train_data = []
            for _ in np.arange(args.n_batch):
                train_data.append(node_sample(randint(), graph, pre_target_nodes, pre_range, args.batch_size))
            valid_data = node_sample(randint(), graph, train_target_nodes, train_range, args.batch_size)

            train_link_losses = []
            train_attr_losses = []
            gpt_gnn.train()
            for data, rem_edge_list, ori_edge_list, attr, (start_idx, end_idx) in train_data:
                node_feature, node_type, edge_time, edge_index, edge_type, node_dict, edge_dict = data
                node_feature = node_feature.detach()
                node_feature[start_idx : end_idx] = gpt_gnn.init_emb
                node_emb = gpt_gnn.gnn(node_feature.to(device), node_type.to(device), edge_time.to(device), \
                                    edge_index.to(device), edge_type.to(device))

                loss_link, _ = gpt_gnn.link_loss(node_emb, rem_edge_list, ori_edge_list, node_dict, target_type, use_queue = True, update_queue=True)
                if args.attr_type == 'text':
                    loss_attr = gpt_gnn.text_loss(node_emb[start_idx : end_idx], attr, w2v_model, device)
                else:
                    loss_attr = gpt_gnn.feat_loss(node_emb[start_idx : end_idx], torch.FloatTensor(attr).to(device))


                loss = loss_link + loss_attr * args.attr_ratio


                optimizer.zero_grad() 
                loss.backward()
                torch.nn.utils.clip_grad_norm_(gpt_gnn.parameters(), args.clip)
                optimizer.step()

                train_link_losses += [loss_link.item()]
                train_attr_losses += [loss_attr.item()]
                scheduler.step()

            # valid 
            gpt_gnn.eval()
            with torch.no_grad():
                data, rem_edge_list, ori_edge_list, attr, (start_idx, end_idx) = valid_data
                node_feature, node_type, edge_time, edge_index, edge_type, node_dict, edge_dict = data
                node_feature = node_feature.detach()
                node_feature[start_idx : end_idx] = gpt_gnn.init_emb
                node_emb = gpt_gnn.gnn(node_feature.to(device), node_type.to(device), edge_time.to(device), \
                                        edge_index.to(device), edge_type.to(device))
                loss_link, ress = gpt_gnn.link_loss(node_emb, rem_edge_list, ori_edge_list, node_dict, target_type, use_queue = False, update_queue=True)
                loss_link = loss_link.item()
                if args.attr_type == 'text':   
                    loss_attr = gpt_gnn.text_loss(node_emb[start_idx : end_idx], attr, w2v_model, device)
                else:
                    loss_attr = gpt_gnn.feat_loss(node_emb[start_idx : end_idx], torch.FloatTensor(attr).to(device))

                ndcgs = []
                for i in ress:
                    ai = np.zeros(len(i[0]))
                    ai[0] = 1
                    ndcgs += [ndcg_at_k(ai[j.cpu().numpy()], len(j)) for j in i.argsort(descending = True)]     
                    
                valid_loss = loss_link + loss_attr * args.attr_ratio
                
                print(("Epoch: %d, (%d / %d)  LR: %.5f Train Loss: (%.3f, %.3f)  Valid Loss: (%.3f, %.3f)  NDCG: %.3f  Norm: %.3f  queue: %d") % \
                    (epoch, batch, repeat_num, optimizer.param_groups[0]['lr'], np.average(train_link_losses), np.average(train_attr_losses), \
                    loss_link, loss_attr, np.average(ndcgs), node_emb.norm(dim=1).mean(), gpt_gnn.neg_queue_size))  
                
            if valid_loss < best_val:
                best_val = valid_loss
                torch.save(gpt_gnn.state_dict(), args.pretrain_model_dir)
            stats += [[np.average(train_link_losses),  loss_link, loss_attr, valid_loss]]


from .gpt_gnn import GPT_GNN, Classifier, Matcher, HGT, RNNModel
from .utils import get_optimizer, LinkPredictor
from .llaga import LLAGA

__all__ = [
    'GPT_GNN',
    'Classifier',
    'Matcher',
    'HGT',
    'RNNModel',
    'get_optimizer',
    'LLAGA',
    'LinkPredictor'
]

classes = __all__



import numpy as np
from collections import defaultdict


def feature_extractor(layer_data, graph):
    r"""`"GPT-GNN: Generative Pre-Training of Graph Neural Networks"
    <https://arxiv.org/abs/2006.15437>`_ paper.

    Extract relevent features.
    
    Parameters
    ----------
    layer_data: dict
        Sampled node indexes for each node type.
    graph: class:`ggfm.data.Graph`
        Target graph.
    """

    feature = {}
    times   = {}
    indxs   = {}
    texts   = []
    for _type in layer_data:
        if len(layer_data[_type]) == 0:
            continue
        idxs  = np.array(list(layer_data[_type].keys()))  #  origin_indxs
        tims  = np.array(list(layer_data[_type].values()))[:,1]  # times
        
        if 'node_emb' in graph.node_feature[_type]:
            feature[_type] = np.array(list(graph.node_feature[_type].loc[idxs, 'node_emb']), dtype=float)
        else:  # intialize as 0
            feature[_type] = np.zeros([len(idxs), 400])
        # 400, 768, 1
        feature[_type] = np.concatenate((feature[_type], list(graph.node_feature[_type].loc[idxs, 'emb']),\
            np.log10(np.array(list(graph.node_feature[_type].loc[idxs, 'citation'])).reshape(-1, 1) + 0.01)), axis=1)
        
        times[_type]   = tims
        indxs[_type]   = idxs
        
        if _type == 'paper':
            attr = np.array(list(graph.node_feature[_type].loc[idxs, 'title']), dtype=str)
    return feature, times, indxs, attr


def sample_subgraph(graph, time_range, sampled_depth = 2, sampled_number = 8, inp=None):
    
    r"""`"GPT-GNN: Generative Pre-Training of Graph Neural Networks"
    <https://arxiv.org/abs/2006.15437>`_ paper.

    Sample Sub-Graph based on the connection of other nodes with currently sampled nodes
    Budgets are maintained for each node type, indexed by <node_id, time>.
    Currently sampled nodes are stored in layer_data.
    After nodes are sampled, the sampled adjacancy matrix are constructed.

    Parameters
    ----------
    graph: class:`ggfm.data.Graph`
        Target graph.
    time_range: list
        Time range of target nodes.
    sampled_depth: int, optional
        Sampled depth.
        (default: :obj:`2`)
    sampled_number: int, optional
        Sampled number.
        (default: :obj:`8`)
    inp: dict
        Input data for sampling. 
        `inp = {target_type: samp_target_nodes}`
    """
    layer_data  = defaultdict( #target_type
                        lambda: {} # {target_id: [ser, time]}
                    )
    budget      = defaultdict( #source_type
                                    lambda: defaultdict(  # source_id
                                        lambda: [0., 0] # [sampled_score, time]
                            ))
    
    '''
        For each node being sampled, we find out all its neighborhood, 
        adding the degree count of these nodes in the budget.
        Note that there exist some nodes that have many neighborhoods
        (such as fields, venues), for those case, we only consider 
    '''
    def add_budget(te, target_id, target_time, layer_data, budget):
        for source_type in te:  # source_type
            tes = te[source_type]  # relation
            for relation_type in tes:  # such as: rev_PV_Conference, rev_PV_Journal
                if relation_type == 'self' or target_id not in tes[relation_type]:
                    continue
                adl = tes[relation_type][target_id]  # {source_id: year, }
                if len(adl) < sampled_number:
                    sampled_ids = list(adl.keys())
                else:
                    sampled_ids = np.random.choice(list(adl.keys()), sampled_number, replace = False)
                for source_id in sampled_ids:
                    source_time = adl[source_id]
                    if source_time == None:
                        source_time = target_time
                    if source_time > np.max(list(time_range.keys())) or source_id in layer_data[source_type]:
                        continue
                    budget[source_type][source_id][0] += 1. / len(sampled_ids)  # score
                    budget[source_type][source_id][1] = source_time  # time

    '''
        First adding the sampled nodes then updating budget.
    '''
    # inp = {target_type: samp_target_nodes}  # inp['paper'].shape = (batch_size, 2) [[id, year], [id, year], ]
    for _type in inp:  # paper
        for _id, _time in inp[_type]:  # id transfer
            layer_data[_type][_id] = [len(layer_data[_type]), _time]  # id -> cur_length // layer_data: {'paper': {id: [cur_length, year], }}
    for _type in inp:  # sampling nodes for each source type of each target type
        te = graph.edge_list[_type]  # such as: paper_venue, paper_paper, paper_field, paper_author
        for _id, _time in inp[_type]:
            add_budget(te, _id, _time, layer_data, budget)
    '''
        We recursively expand the sampled graph by sampled_depth.
        Each time we sample a fixed number of nodes for each budget,
        based on the accumulated degree.
    '''
    for layer in range(sampled_depth):
        sts = list(budget.keys())
        for source_type in sts:
            te = graph.edge_list[source_type]
            keys = np.array(list(budget[source_type].keys()))
            if sampled_number > len(keys):
                sampled_ids = np.arange(len(keys))
            else:
                score = np.array(list(budget[source_type].values()))[:,0] ** 2
                score = score / np.sum(score)
                sampled_ids = np.random.choice(len(score), sampled_number, p = score, replace = False) 
            sampled_keys = keys[sampled_ids]
            '''
                First adding the sampled nodes then updating budget.
            '''
            for k in sampled_keys:
                layer_data[source_type][k] = [len(layer_data[source_type]), budget[source_type][k][1]]  # layer_data[source_type] {id: [cur_length, time], }
            for k in sampled_keys:
                add_budget(te, k, budget[source_type][k][1], layer_data, budget)
                budget[source_type].pop(k)
    
    # Prepare feature, time and adjacency matrix for the sampled graph, indxs are the origin indexes, texts are the title information of papers
    feature, times, indxs, texts = feature_extractor(layer_data, graph)
            
    edge_list = defaultdict(  # target_type
                        lambda: defaultdict(  # source_type
                            lambda: defaultdict(  # relation_type
                                lambda: [] # [target_id, source_id] 
                                    )))
    for _type in layer_data:  # {type: {id: [cur_id, year], }, }
        for _key in layer_data[_type]:
            _ser = layer_data[_type][_key][0]  # cur_id
            edge_list[_type][_type]['self'] += [[_ser, _ser]]  # add self-loop
    
    '''
        Reconstruct sampled adjacancy matrix by checking whether each
        link exist in the original graph
    '''
    for target_type in graph.edge_list:
        te = graph.edge_list[target_type]
        tld = layer_data[target_type]  # {type: {id: [cur_id, year], }, }
        for source_type in te:
            tes = te[source_type]
            sld  = layer_data[source_type]
            for relation_type in tes:  # relation
                tesr = tes[relation_type]  # target_id
                for target_key in tld:  # sampled target_ids
                    if target_key not in tesr:
                        continue
                    target_ser = tld[target_key][0]  # cur_id
                    for source_key in tesr[target_key]:
                        # Check whether each link (target_id, source_id) exist in original adjacancy matrix
                        if source_key in sld:
                            source_ser = sld[source_key][0]
                            edge_list[target_type][source_type][relation_type] += [[target_ser, source_ser]]
    
    return feature, times, edge_list, indxs, texts


import pickle
import numpy as np
from texttable import Texttable

def open_pkl_file(file_path):
    r"""
    Open pickle file.

    Parameters
    ----------
    file_path: str
        File path for loading pickle files.
    """
    with open(file_path, 'rb') as file:
        file_content = pickle.load(file)
        return file_content


def save_pkl_file(file_path, contents):
    r"""
    Save pickle file.

    Parameters
    ----------
    file_path: str
        File path for saving pickle files.
    contents: list
        Contents for saving.
    """
    with open(file_path, 'wb') as file:
        pickle.dump(contents, file)
    print("having saved pkl...")


def open_txt_file(file_path):
    r"""
    Open txt file.

    Parameters
    ----------
    file_path: str
        File path for loading txt files.
    """
    with open(file_path, 'r') as file:
        contents = [line.rstrip("\n") for line in file.readlines()]
        return contents


def save_txt_file(file_path, contents):
    r"""
    Save txt file.

    Parameters
    ----------
    file_path: str
        File path for saving txt files.
    contents: list
        Contents for saving.
    """
    with open(file_path, 'w') as file:
        for paragraph in contents:
            file.write(paragraph + "\n")
    print("having saved txt...")


def dcg_at_k(r, k):
    r = np.asfarray(r)[:k]
    if r.size:
        return r[0] + np.sum(r[1:] / np.log2(np.arange(2, r.size + 1)))
    return 0.


def ndcg_at_k(r, k):
    r"""

    Compute the Normalized Discounted Cumulative Gain (NDCG) at rank k.

    Parameters
    ----------
    r: list
        A list of relevance scores representing the ranking of items.
    k: int
        The rank at which to compute NDCG.
    
    Returns
    -------
    float
      The Normalized Discounted Cumulative Gain (NDCG) value.
    """
    dcg_max = dcg_at_k(sorted(r, reverse=True), k)
    if not dcg_max:
        return 0.
    return dcg_at_k(r, k) / dcg_max


def mean_reciprocal_rank(rs):
    r"""
    Compute the Mean Reciprocal Rank (MRR) for a list of relevance scores.

    Parameters
    ----------
    rs: list of arrays
        A list of relevance score arrays where each array represents the indices of relevant items.

    Returns
    -------
    list
        A list of MRR values for each query.
    """
    rs = (np.asarray(r).nonzero()[0] for r in rs)
    return [1. / (r[0] + 1) if r.size else 0. for r in rs]


def args_print(args):
    r"""
    Print argments.

    Parameters
    ----------
    args: object
        args
    """
    _dict = vars(args)
    t = Texttable() 
    t.add_row(["Parameter", "Value"])
    for k in _dict:
        t.add_row([k, _dict[k]])
    print(t.draw())


import torch
import numpy as np


def generate_lm_embs(all_name, tokenizer, model, indxs, device):
    r"""

    Language model generates node embeddings for words。

    Parameters
    ----------
    all_name: list
        Words list.

    tokenizer: 
        LM tokenizer.
    
    model: 
        Language model.
    
    indxs: list
        Index positions of generated node embeddings.
    
    device: int
        Device.


    """

    def get_word_embeddings(word, device):
        encoded_word = tokenizer.encode(word, add_special_tokens=False)
        tokens_tensor = torch.tensor([encoded_word]).to(device)
        with torch.no_grad():
            output = model(tokens_tensor)
            embeddings = output[0][0].mean(dim=0)
        return embeddings.cpu().numpy()

    emb = get_word_embeddings("hello", device)
    emb = np.zeros((len(indxs), len(emb)))

    for i in range(len(indxs)):
        cur = indxs[i]
        word = all_name[cur]
        emb[i] = get_word_embeddings(word, device)

    return emb


import torch
import torch.nn as nn
import torch.nn.functional as F

def get_optimizer(parameters, name, optimizer_args):
        
    if name == "adam":
        optimizer = torch.optim.Adam(parameters, **optimizer_args)
    elif name == "adamw":
        optimizer = torch.optim.AdamW(parameters, **optimizer_args)
    elif name == "adadelta":
        optimizer = torch.optim.Adadelta(parameters, **optimizer_args)
    elif name == "radam":
        optimizer = torch.optim.RAdam(parameters, **optimizer_args)
    else:
        return NotImplementedError
    
    return optimizer

class LinkPredictor(nn.Module):
    r"""LinkPredictor for graph link prediction task.

    Parameters
    ----------
    n_hid: int
        Input size.
    n_out: int
        Output size.

    """
    def __init__(self, n_hid, n_out):
        super(LinkPredictor, self).__init__()

        self.fc1 = nn.Linear(n_hid * 2, n_hid)
        self.fc2 = nn.Linear(n_hid, n_out)

    def forward(self, src, dst):
        x = torch.cat([src, dst], 1)
        y = self.fc2(F.relu(self.fc1(x)))
        return y


import time
import random
from sklearn.model_selection import train_test_split

from warnings import filterwarnings
filterwarnings("ignore")


def construct_link_and_node(graph, data_dir):
    r"""

    Construct link.dat and node.dat.

    Parameters
    ----------
    graph: class:`ggfm.data.Graph`
        Target graph.
    
    data_dir: str
        Data directory for saving link.dat and node.dat, which are saved as data_dir/link.dat and data_dir/node.dat.
    """

    type_begin_ids, graph_node_type = get_type_id(graph)
    
    graph_node_name = {}
    for i in range(len(graph_node_type)):
        attr = "name"
        if graph_node_type[i] == "paper": attr = "title"
        graph_node_name[graph_node_type[i]] = graph.node_feature[graph_node_type[i]][attr].tolist()

    # node.dat
    node_type2id = {}
    type_num = 0
    for i in range(len(graph_node_type)):
        node_type2id[graph_node_type[i]] = type_num
        type_num += 1
    
    with open(data_dir + "node.dat", "w") as file:
        for i in range(len(graph_node_type)):
            for j in range(len(graph_node_name[graph_node_type[i]])):
                node_id = j + type_begin_ids[i]
                node_name = graph_node_name[graph_node_type[i]][j]
                node_type = node_type2id[graph_node_type[i]]
                file.write(f"{node_id}\t{node_name}\t{node_type}\n")
    
    print("node.dat has been saved.")
        
    
    # link.dat
    # start_id  end_id  edge_type
    edges = graph.edge_list
    relations = graph.get_meta_graph()
    edge_types = {}
    num_edge = 0
    for r in relations:
        edge_types[r[2]] = num_edge
        num_edge += 1
    
    with open(data_dir + "link.dat", "w") as file:
        for target_type in edges:
            for source_type in edges[target_type]:
                for relation_type in edges[target_type][source_type]:
                    for target_id in edges[target_type][source_type][relation_type]:
                        for source_id in edges[target_type][source_type][relation_type][target_id]:
                            src = source_id + type_begin_ids[node_type2id[source_type]]
                            dst = target_id + type_begin_ids[node_type2id[target_type]]
                            edge_type = edge_types[relation_type]
                            file.write(f"{src}\t{dst}\t{edge_type}\n")

    print("link.dat has been saved.")


def get_type_id(graph):
    r"""

    Statistically analyze the type_degin_ids and graph_node_type of graphs.

    Parameters
    ----------
    graph: class:`ggfm.data.Graph`
        Target graph.
    
    Returns
    -------
    type_begin_ids: list
        The begin index of each node type, type_begin_ids are consistent with graph_node_type.
    graph_node_type: list
        Graph node types.
    """

    graph_node_name = {}
    graph_node_type = graph.get_types()
    for i in range(len(graph_node_type)):
        attr = "name"
        if graph_node_type[i] == "paper": attr = "title"
        graph_node_name[graph_node_type[i]] = graph.node_feature[graph_node_type[i]][attr].tolist()

    type_begin_ids = [0, ]
    for i in range(1, len(graph_node_type)):
        type_begin_ids.append(type_begin_ids[i-1]+len(graph_node_name[graph_node_type[i-1]]))
    
    return type_begin_ids, graph_node_type



def random_walk_based_corpus_construction(data_dir, relations, alpha=0.05, path_length=1000000, path_num=450000):
    r"""

    Construct link.dat and node.dat.

    Parameters
    ----------
    data_dir: str
        Data directory for loading link.dat and node.dat, also for saving output.txt, rw_train_corpus.txt and rw_valid_corpus.txt.
    
    relations: list
        Relations for all edge types.
    
    alpha: str, optional
        Each path will terminate sampling with a probability of alpha.
        (default: :obj:`0.05`)
    
    path_length: int, optional
        Sampling length of each path.
        (default: :obj:`1000000`)
    
    path_num: int, optional
        Number of sampled paths.
        (default: :obj:`450000`)

    """

    op1, op2, op3 = [], [], []

    with open(data_dir + 'node.dat','r') as file:
        for line in file:
            node_id, node_name, node_type = line.split('\t')
            op1.append(int(node_id))
            op2.append(node_name)
            op3.append(int(node_type))  
        

    G=[[] for i in range(len(op3))]

    with open(data_dir + 'link.dat', 'r') as file:
        for line in file:
            src, dst, edge_type = line.split('\t')
            G[int(src)].append([int(dst), int(edge_type)])

    line_idx = op1
    rand = random.Random()
    patient_patient_path = []

    dic = {}
    start_time = time.time()
    for line in range(path_num):
        if line % 10000 == 0:
            current_time = time.time()
            dual_time = current_time - start_time
            print(f"having sampling {line} lines and spent {dual_time}...")
        temp_path = []
        start_path = rand.choice(line_idx)
        temp_path.append([start_path,-1])
        dic[start_path] = 1
        for i in range(path_length):
            cur = temp_path[-1][0]
            if (len(G[cur]) > 0):
                if rand.random() >= alpha:
                    cur_path = rand.choice(G[cur])
                    temp_path.append(cur_path)
                    dic[cur_path[0]] = 1
                else:
                    break
            else:
                break
        if (len(temp_path) >= 2):
            patient_patient_path.append(temp_path)


    line_name = {}
    for i in range(len(relations)):
        line_name[i] = relations[i]


    with open(data_dir + 'output.txt', 'w') as f:
        for i in range(len(patient_patient_path)):
            print(op2[patient_patient_path[i][0][0]],line_name[patient_patient_path[i][1][1]],op2[patient_patient_path[i][1][0]],end='',file=f)
            for j in range(1,len(patient_patient_path[i])-2):
                print(' '+line_name[patient_patient_path[i][j+1][1]],op2[patient_patient_path[i][j+1][0]],end='',file=f)
            if(len(patient_patient_path[i])>2):
                print(' '+line_name[patient_patient_path[i][-1][1]],op2[patient_patient_path[i][-1][0]],end='',file=f)
            print("\n",end='',file=f)


    with open(data_dir + 'output.txt', 'r') as file:
        corpus = [line.rstrip("\n") for line in file.readlines()]
    
    print(f"length of corpus: {len(corpus)}")
    print("corpus[0]: ")
    print(corpus[0])

    train_text, val_text = train_test_split(corpus, test_size=0.15, random_state=42)

    with open(data_dir + 'rw_train_corpus.txt', 'w') as file:
        for paragraph in train_text:
            file.write(paragraph + "\n")
            
    with open(data_dir + 'rw_val_corpus.txt', 'w') as file:
        for paragraph in val_text:
            file.write(paragraph + "\n")

    print(f"length of train_corpus: {len(train_text)}")
    print(f"length of valid_corpus: {len(val_text)}")
    print("train_corpus and valid_corpus have been saved.")


import dill
from collections import defaultdict


class Graph():
    r"""The Heterogeneous Graph Transformer (HGT) operator from the
    `"Heterogeneous Graph Transformer" 
    <https://arxiv.org/abs/2003.01332>`_ paper.
    
    A Graph object describe a homogeneous/heterogeneous graph. 
    The node-related information is included in node_feature: a DataFrame containing all features
    The link-related information is included in edge_list: index the adjacancy matrix (time) by 
        <target_type, source_type, relation_type, target_id, source_id>
    """

    def __init__(self):
        super(Graph, self).__init__()
        '''
            node_forward and bacward are only used when building the data. 
            Afterwards will be transformed into node_feature by DataFrame
            
            node_forward: name -> node_id
            node_bacward: node_id -> feature_dict
            node_feature: a DataFrame containing all features
        '''
        self.node_forward = defaultdict(lambda: {})
        self.node_bacward = defaultdict(lambda: [])
        self.node_feature = defaultdict(lambda: [])

        '''
            edge_list: index the adjacancy matrix (time) by 
            <target_type, source_type, relation_type, target_id, source_id>
        '''
        self.edge_list = defaultdict( #target_type
                            lambda: defaultdict(  #source_type
                                lambda: defaultdict(  #relation_type
                                    lambda: defaultdict(  #target_id
                                        lambda: defaultdict( # source_id
                                            lambda: int # time
                                        )))))
        self.times = {}
    def add_node(self, node):
        nfl = self.node_forward[node['type']]
        if node['id'] not in nfl:
            self.node_bacward[node['type']] += [node]  # type: [node(dict), node(dict),]
            ser = len(nfl)
            nfl[node['id']] = ser  # id -> cur_length
            return ser
        return nfl[node['id']]
    def add_edge(self, source_node, target_node, time = None, relation_type = None, directed = True):
        edge = [self.add_node(source_node), self.add_node(target_node)]
        '''
            Add bi-directional edges with different relation type
        '''
        self.edge_list[target_node['type']][source_node['type']][relation_type][edge[1]][edge[0]] = time
        if directed:
            self.edge_list[source_node['type']][target_node['type']]['rev_' + relation_type][edge[0]][edge[1]] = time
        else:
            self.edge_list[source_node['type']][target_node['type']][relation_type][edge[0]][edge[1]] = time
        self.times[time] = True
        
    def update_node(self, node):
        nbl = self.node_bacward[node['type']]
        ser = self.add_node(node)
        for k in node:
            if k not in nbl[ser]:
                nbl[ser][k] = node[k]

    def get_meta_graph(self):
        types = self.get_types()
        metas = []
        for target_type in self.edge_list:
            for source_type in self.edge_list[target_type]:
                for r_type in self.edge_list[target_type][source_type]:
                    metas += [(target_type, source_type, r_type)]
        return metas
    
    def get_types(self):
        return list(self.node_feature.keys())


class RenameUnpickler(dill.Unpickler):
    def find_class(self, module, name):
        renamed_module = module
        if module == "data":
            renamed_module = "ggfm.data.graph"
        return super(RenameUnpickler, self).find_class(renamed_module, name)


def renamed_load(file_obj):
    r"""
        Load Graph object.
    """
    return RenameUnpickler(file_obj).load()



import os
import torch
import argparse
import numpy as np
import torch.nn as nn
from ggfm.data import args_print
from collections import OrderedDict
from sklearn.metrics import f1_score
from ggfm.models import HGT, Classifier
from ggfm.data import renamed_load, sample_subgraph, open_pkl_file

from warnings import filterwarnings
filterwarnings("ignore")


def to_torch(feature, time, edge_list, graph):
    '''
        Transform a sampled sub-graph into pytorch Tensor
        node_dict: {node_type: <node_number, node_type_ID>} node_number is used to trace back the nodes in original graph.
        edge_dict: {edge_type: edge_type_ID}
    '''
    node_dict = {}
    node_feature = []
    node_type    = []
    node_time    = []
    edge_index   = []
    edge_type    = []
    edge_time    = []
    
    node_num = 0
    types = graph.get_types()
    for t in types:
        node_dict[t] = [node_num, len(node_dict)]
        node_num += len(feature[t])
        
    for t in types:
        node_feature += list(feature[t])
        node_time    += list(time[t])
        node_type    += [node_dict[t][1] for _ in range(len(feature[t]))]
        
    edge_dict = {e[2]: i for i, e in enumerate(graph.get_meta_graph())}
    edge_dict['self'] = len(edge_dict)

    # transfer to homogeneous graph
    for target_type in edge_list:
        for source_type in edge_list[target_type]:
            for relation_type in edge_list[target_type][source_type]:
                for ii, (ti, si) in enumerate(edge_list[target_type][source_type][relation_type]):
                    tid, sid = ti + node_dict[target_type][0], si + node_dict[source_type][0]
                    edge_index += [[sid, tid]]
                    edge_type  += [edge_dict[relation_type]]

                    # time ranges from 1900 - 2020, largest span is 120.
                    edge_time  += [node_time[tid] - node_time[sid] + 120]

    node_feature = torch.FloatTensor(node_feature)
    node_type    = torch.LongTensor(node_type)
    edge_time    = torch.LongTensor(edge_time)
    edge_index   = torch.LongTensor(edge_index).t()
    edge_type    = torch.LongTensor(edge_type)
    return node_feature, node_type, edge_time, edge_index, edge_type, node_dict, edge_dict


def node_sample(seed, graph, args, pairs, time_range, output_num):
    '''
        sub-graph sampling and label preparation for node classification:
        (1) Sample batch_size number of output nodes (papers), get their time.
    '''
    np.random.seed(seed)
    target_ids = np.random.choice(list(pairs.keys()), args.batch_size, replace = False)
    target_info = []
    for target_id in target_ids:
        _, _time = pairs[target_id]
        target_info += [[target_id, _time]]
    '''
        (2) Based on the seed nodes, sample a subgraph with 'sampled_depth' and 'sampled_number'
    '''
    feature, times, edge_list, _, _ = sample_subgraph(graph, time_range, inp = {'paper': np.array(target_info)}, sampled_depth = args.sample_depth, sampled_number = args.sample_width)

    '''
        (3) Mask out the edge between the output target nodes (paper) with output source nodes (L2 field)
    '''
    masked_edge_list = []
    for i in edge_list['paper']['field']['rev_PF_in_L2']:
        if i[0] >= args.batch_size:  # only save the non mask-idxs
            masked_edge_list += [i]
    edge_list['paper']['field']['rev_PF_in_L2'] = masked_edge_list

    masked_edge_list = []
    for i in edge_list['field']['paper']['PF_in_L2']:
        if i[1] >= args.batch_size:
            masked_edge_list += [i]
    edge_list['field']['paper']['PF_in_L2'] = masked_edge_list
    '''
        (4) Transform the subgraph into torch Tensor (edge_index is in format of pytorch_geometric)
    '''
    node_feature, node_type, edge_time, edge_index, edge_type, node_dict, edge_dict = to_torch(feature, times, edge_list, graph)
    '''
        (5) Prepare the labels for each output target node (paper), and their index in sampled graph.
            (node_dict[type][0] stores the start index of a specific type of nodes)
    '''
    cand_list = list(graph.edge_list['field']['paper']['PF_in_L2'].keys())  # field_id

    ylabel = np.zeros([args.batch_size, output_num])
    for x_id, target_id in enumerate(target_ids):
        if target_id not in pairs:
            print('error 1' + str(target_id))
        for source_id in pairs[target_id][0]:
            if source_id not in cand_list:
                print('error 2' + str(target_id))
            ylabel[x_id][field_labels[source_id]] = 1

    ylabel /= ylabel.sum(axis=1).reshape(-1, 1)
    x_ids = np.arange(args.batch_size) + node_dict['paper'][0]
    return node_feature, node_type, edge_time, edge_index, edge_type, x_ids, ylabel


def load_gnn(_dict):
    out_dict = {}
    for key in _dict:
        if 'gnn' in key:
            out_dict[key[4:]] = _dict[key]
    return OrderedDict(out_dict)


def randint():
    return np.random.randint(2**32 - 1)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Fine-Tuning on OAG Paper-Field (L2) classification task')

    '''
        Dataset arguments
    '''
    parser.add_argument('--data_dir', type=str, default='/home/yjy/heteroPrompt/ggfm/ggfm/datasets/', help='The address of data.')
    parser.add_argument('--use_pretrain', type=str, default=True, help='Whether to use pre-trained model')
    parser.add_argument('--pretrain_model_dir', type=str, default='/home/yjy/heteroPrompt/ggfm/ggfm/pretrained_model/gta_all_cs3', help='The address for pretrained model.')
    parser.add_argument('--model_dir', type=str, default='/home/yjy/heteroPrompt/ggfm/ggfm/fine_tuned_model', help='The address for storing the models and optimization results.')
    parser.add_argument('--task_name', type=str, default='gptgnn_nc', help='The name of the stored models and optimization results.')
    parser.add_argument('--cuda', type=int, default=0, help='Avaiable GPU ID')
    parser.add_argument('--sample_depth', type=int, default=6, help='How many numbers to sample the graph')
    parser.add_argument('--sample_width', type=int, default=128, help='How many nodes to be sampled per layer per type')

    '''
    Model arguments 
    '''
    parser.add_argument('--conv_name', type=str, default='hgt', help='The name of GNN filter. By default is Heterogeneous Graph Transformer (hgt)')
    parser.add_argument('--n_hid', type=int, default=400, help='Number of hidden dimension')
    parser.add_argument('--n_heads', type=int, default=8, help='Number of attention head')
    parser.add_argument('--n_layers', type=int, default=3, help='Number of GNN layers')
    parser.add_argument('--prev_norm', help='Whether to add layer-norm on the previous layers', action='store_true')
    parser.add_argument('--last_norm', help='Whether to add layer-norm on the last layers', action='store_true')
    parser.add_argument('--dropout', type=int, default=0.2, help='Dropout ratio')


    '''
        Optimization arguments
    '''
    parser.add_argument('--scheduler', type=str, default='cycle', help='Name of learning rate scheduler.' , choices=['cycle', 'cosine'])
    parser.add_argument('--data_percentage', type=int, default=0.1, help='Percentage of training and validation data to use')
    parser.add_argument('--n_epoch', type=int, default=50, help='Number of epoch to run')
    parser.add_argument('--n_pool', type=int, default=8, help='Number of process to sample subgraph')    
    parser.add_argument('--n_batch', type=int, default=16, help='Number of batch (sampled graphs) for each epoch')
    parser.add_argument('--batch_size', type=int, default=256, help='Number of output nodes for training')
    parser.add_argument('--clip', type=int, default=0.5, help='Gradient Norm Clipping')

    args = parser.parse_args()
    args_print(args)

    if args.cuda != -1: device = torch.device("cuda:" + str(args.cuda))
    else: device = torch.device("cpu")


    # pre data for fine-tuning
    graph = renamed_load(open(args.data_dir+"graph.pk", 'rb'))

    target_type = 'paper'
    types = graph.get_types()

    pre_range   = {t: True for t in graph.times if t != None and t < 2014}
    train_range = {t: True for t in graph.times if t != None and t >= 2014 and t <= 2016}
    valid_range = {t: True for t in graph.times if t != None and t > 2016 and t <= 2017}
    test_range  = {t: True for t in graph.times if t != None and t > 2017}


    train_ids = open_pkl_file(args.data_dir+"train_ids.pkl")
    valid_ids = open_pkl_file(args.data_dir+"valid_ids.pkl")
    test_ids = open_pkl_file(args.data_dir+"test_ids.pkl")

    # get pairs for labels
    train_pairs = {}
    valid_pairs = {}
    test_pairs  = {}
    
    # Prepare all the souce nodes (L2 field) associated with each target node (paper) as dict
    for target_id in graph.edge_list['paper']['field']['rev_PF_in_L2']:  # paper_id
        for source_id in graph.edge_list['paper']['field']['rev_PF_in_L2'][target_id]:  # field_id
            _time = graph.edge_list['paper']['field']['rev_PF_in_L2'][target_id][source_id]  # time
            if target_id in train_ids:
                if target_id not in train_pairs:
                    train_pairs[target_id] = [[], _time]
                train_pairs[target_id][0] += [source_id]

            elif target_id in valid_ids:
                if target_id not in valid_pairs:
                    valid_pairs[target_id] = [[], _time]
                valid_pairs[target_id][0] += [source_id]
            else:
                if target_id not in test_pairs:
                    test_pairs[target_id]  = [[], _time]
                test_pairs[target_id][0]  += [source_id]

    
    # Only train and valid with a certain percentage of data, if necessary.
    np.random.seed(43)
    sel_train_pairs = {p : train_pairs[p] for p in np.random.choice(list(train_pairs.keys()), int(len(train_pairs) * args.data_percentage), replace = False)}
    sel_valid_pairs = {p : valid_pairs[p] for p in np.random.choice(list(valid_pairs.keys()), int(len(valid_pairs) * args.data_percentage), replace = False)}


    field_labels = list(graph.node_feature['field']['label'])
    output_num = max(field_labels) + 1  # 11 paper node classification


    # Initialize GNN (model is specified by conv_name) and Classifier
    gnn = HGT(in_dim = len(graph.node_feature[target_type]['emb'].values[0]) + 401, n_hid = args.n_hid, \
            n_heads = args.n_heads, n_layers = args.n_layers, dropout = args.dropout, num_types = len(types), \
            num_relations = len(graph.get_meta_graph()) + 1, prev_norm = args.prev_norm, last_norm = args.last_norm)
    if args.use_pretrain:
        gnn.load_state_dict(load_gnn(torch.load(args.pretrain_model_dir)), strict = False)
        print('Load Pre-trained Model from (%s)' % args.pretrain_model_dir)
    classifier = Classifier(args.n_hid, output_num)
    model = nn.Sequential(gnn, classifier).to(device)

    criterion = nn.KLDivLoss(reduction='batchmean')
    optimizer_args = dict(lr=5e-4)
    optimizer = torch.optim.AdamW(model.parameters(), **optimizer_args)
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, 500, eta_min=1e-6)

    res = []
    best_val = 0
    train_step = 0
    
    for epoch in np.arange(args.n_epoch) + 1:
        train_data = []
        for _ in np.arange(args.n_batch):
            train_data.append(node_sample(randint(), graph, args, sel_train_pairs, train_range, output_num))
        valid_data = node_sample(randint(), graph, args, sel_valid_pairs, valid_range, output_num)
        
        
        model.train()
        train_losses = []
        for node_feature, node_type, edge_time, edge_index, edge_type, x_ids, ylabel in train_data:
            node_rep = gnn.forward(node_feature.to(device), node_type.to(device), edge_time.to(device), edge_index.to(device), edge_type.to(device))
            res  = classifier.forward(node_rep[x_ids])
            loss = criterion(res, torch.FloatTensor(ylabel).to(device))

            optimizer.zero_grad() 
            loss.backward()

            torch.nn.utils.clip_grad_norm_(model.parameters(), args.clip)
            optimizer.step()

            train_losses += [loss.cpu().detach().tolist()]
            train_step += 1
            scheduler.step(train_step)
            del res, loss
        
        model.eval()
        with torch.no_grad():
            final_preds, final_labels = [], []

            node_feature, node_type, edge_time, edge_index, edge_type, x_ids, ylabel = valid_data
            node_rep = gnn.forward(node_feature.to(device), node_type.to(device), edge_time.to(device), edge_index.to(device), edge_type.to(device))
            res = classifier.forward(node_rep[x_ids])
            loss = criterion(res, torch.FloatTensor(ylabel).to(device))

            predictions, ground_truth = [], []

            output = res.detach().cpu()
            labels = torch.from_numpy(ylabel)

            cates = torch.count_nonzero(labels, dim=1).tolist()

            for i in range(output.shape[0]):
                k = cates[i]
                _, pred_indxs = torch.topk(output[i], k)
                predictions.append(pred_indxs.tolist())
                _, label_indxs = torch.topk(labels[i], k)
                ground_truth.append(label_indxs.tolist())
            
            for i in range(len(predictions)):
                flag = 0
                predictions[i] = sorted(predictions[i])
                ground_truth[i] = sorted(ground_truth[i])
                for j in range(len(predictions[i])):
                    if predictions[i][j] != ground_truth[i][j]:
                        flag = 1
                        break
                if flag == 1: final_preds.append(0)
                else: final_preds.append(1)
                final_labels.append(1)

            ma = f1_score(final_labels, final_preds, average='macro')
            mi = f1_score(final_labels, final_preds, average='micro')

            if mi > best_val:
                best_val = mi
                torch.save(model, os.path.join(args.model_dir, args.task_name + '_' + args.conv_name))

            print(("Epoch: %d  LR: %.5f Train Loss: %.2f  Valid Loss: %.2f  Valid Micro-F1: %.4f  Valid Macro-F1: %.4f") % \
                  (epoch, optimizer.param_groups[0]['lr'], np.average(train_losses), loss.cpu().detach().tolist(), mi, ma))
            del res, loss
        del train_data, valid_data

    
    best_model = torch.load(os.path.join(args.model_dir, args.task_name + '_' + args.conv_name))
    best_model.eval()
    gnn, classifier = best_model
    with torch.no_grad():
        final_preds, final_labels = [], []

        test_res = []
        all_res = []
        
        node_feature, node_type, edge_time, edge_index, edge_type, x_ids, ylabel = node_sample(randint(), graph, args, test_pairs, test_range, output_num)
        paper_rep = gnn.forward(node_feature.to(device), node_type.to(device), edge_time.to(device), edge_index.to(device), edge_type.to(device))[x_ids]
        res = classifier.forward(paper_rep)
        
        predictions, ground_truth = [], []

        output = res.detach().cpu()
        labels = torch.from_numpy(ylabel)

        cates = torch.count_nonzero(labels, dim=1).tolist()

        for i in range(output.shape[0]):
            k = cates[i]
            _, pred_indxs = torch.topk(output[i], k)
            predictions.append(pred_indxs.tolist())
            _, label_indxs = torch.topk(labels[i], k)
            ground_truth.append(label_indxs.tolist())
        
        for i in range(len(predictions)):
            flag = 0
            predictions[i] = sorted(predictions[i])
            ground_truth[i] = sorted(ground_truth[i])
            for j in range(len(predictions[i])):
                if predictions[i][j] != ground_truth[i][j]:
                    flag = 1
                    break
            if flag == 1: final_preds.append(0)
            else: final_preds.append(1)
            final_labels.append(1)

        ma = f1_score(final_labels, final_preds, average='macro')
        mi = f1_score(final_labels, final_preds, average='micro')

        print(f"Test Micro-F1: {mi}, Test Macro-F1: {ma}")
